# -*- generated by 1.0.12 -*-
import da
_config_object = {}

class Node_(da.NodeProcess):

    def __init__(self, procimpl, props):
        super().__init__(procimpl, props)
        self._events.extend([])

    def run(self):
        return 0
